from kubemq.grpc import Result


class EventSendResult:
    """
    Initializes an EventSendResult instance.

    Args:
        id (str, optional): The ID of the event. Defaults to None.
        sent (bool, optional): Indicates whether the event was sent successfully. Defaults to False.
        error (str, optional): An error message related to the event send operation. Defaults to None.
    """
    def __init__(self, id: str = None,
                 sent: bool = False,
                 error: str = None):
        self.id: str = id
        self.sent: bool = sent
        self.error: str = error

    def decode(self, result: Result) -> 'EventSendResult':
        self.id = result.EventID
        self.sent = result.Sent
        self.error = result.Error
        return self

    def __repr__(self):
        return f"EventSendResult: id={self.id}, sent={self.sent}, error={self.error}"



from typing import Callable
from kubemq.grpc import Subscribe
from kubemq.common.subscribe_type import SubscribeType
from kubemq.pubsub import EventMessageReceived


class EventsSubscription:
    """
    Class representing an event subscription.

    Args:
        channel (str, optional): The channel to subscribe to. Defaults to None.
        group (str, optional): The group to subscribe to. Defaults to None.
        on_receive_event_callback (Callable[[EventMessageReceived], None], optional):
            The callback function to be called when an event is received. Defaults to None.
        on_error_callback (Callable[[str], None], optional):
            The callback function to be called when an error occurs. Defaults to None.

    Attributes:
        channel (str): The channel to subscribe to.
        group (str): The group to subscribe to.
        on_receive_event_callback (Callable[[EventMessageReceived], None]):
            The callback function to be called when an event is received.
        on_error_callback (Callable[[str], None]):
            The callback function to be called when an error occurs.

    Methods:
        raise_on_receive_message(received_event: EventMessageReceived):
            Raises the on_receive_event_callback with the received event if it is defined.
        raise_on_error(msg: str):
            Raises the on_error_callback with the specified error message if it is defined.
        validate():
            Validates that the event subscription has a channel and an on_receive_event_callback function.
        encode(client_id: str = "") -> Subscribe:
            Encodes the event subscription into a Subscribe object.
        __repr__() -> str:
            Returns a string representation of the EventsSubscription.

    Raises:
        ValueError: If the event subscription does not have a channel or an on_receive_event_callback function.
    """
    def __init__(self, channel: str = None,
                 group: str = None,
                 on_receive_event_callback: Callable[[EventMessageReceived], None] = None,
                 on_error_callback: Callable[[str], None] = None):
        self.channel: str = channel
        self.group: str = group
        self.on_receive_event_callback = on_receive_event_callback
        self.on_error_callback = on_error_callback



    def raise_on_receive_message(self, received_event: EventMessageReceived):
        if self.on_receive_event_callback:
            self.on_receive_event_callback(received_event)

    def raise_on_error(self, msg: str):
        if self.on_error_callback:
            self.on_error_callback(msg)

    def validate(self):
        if not self.channel:
            raise ValueError("Event subscription must have a channel.")
        if not self.on_receive_event_callback:
            raise ValueError("Event subscription must have a on_receive_event_callback function.")

    def encode(self, client_id: str = "") -> Subscribe:
        request = Subscribe()
        request.Channel = self.channel
        request.Group = self.group
        request.ClientID = client_id
        request.SubscribeTypeData = SubscribeType.Events.value
        return request

    def __repr__(self):
        return f"EventsSubscription: channel={self._channel}, group={self._group}"


import uuid
from typing import Callable
from kubemq.grpc import QueuesDownstreamRequest, QueuesDownstreamRequestType
from kubemq.queues.queues_message_received import QueueMessageReceived


class QueuesPollRequest:
    """
    Class representing a request to poll messages from a queue.

    Attributes:
        channel (str): The channel to subscribe to.
        poll_max_messages (int): The maximum number of messages to poll in a single request.
        poll_wait_timeout_in_seconds (int): The maximum time to wait for messages in seconds.
        auto_ack_messages (bool): Whether to automatically acknowledge received messages.

    Methods:
        validate(): Validates the request parameters.
        encode(client_id: str) -> QueuesDownstreamRequest: Encodes the request into a downstream request object.
        __repr__(): Returns a string representation of the QueuesPollRequest object.
    """
    def __init__(self, channel: str = None,
                 poll_max_messages: int = 1,
                 poll_wait_timeout_in_seconds: int = 60,
                 auto_ack_messages: bool = False,
                 ):
        self.channel: str = channel
        self.poll_max_messages: int = poll_max_messages
        self.poll_wait_timeout_in_seconds: int = poll_wait_timeout_in_seconds
        self.auto_ack_messages: bool = auto_ack_messages

    def validate(self):
        if not self.channel:
            raise ValueError("queue subscription must have a channel.")
        if self.poll_max_messages < 1:
            raise ValueError("queue subscription poll_max_messages must be greater than 0.")
        if self.poll_wait_timeout_in_seconds < 1:
            raise ValueError("queue subscription poll_wait_timeout_in_seconds must be greater than 0.")

    def encode(self, client_id: str = "") -> QueuesDownstreamRequest:
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = client_id
        request.Channel = self.channel
        request.MaxItems = self.poll_max_messages
        request.WaitTimeout = self.poll_wait_timeout_in_seconds * 1000
        request.AutoAck = self.auto_ack_messages
        request.RequestTypeData = QueuesDownstreamRequestType.Get
        return request

    def __repr__(self):
        return f"QueuesSubscription: channel={self.channel}, poll_max_messages={self.poll_max_messages}, " \
               f"poll_wait_timeout_in_seconds={self.poll_wait_timeout_in_seconds}, auto_ack_messages={self.auto_ack_messages}"


from datetime import datetime
from typing import Dict
from kubemq.cq.query_message_received import QueryMessageReceived
from kubemq.grpc import Response as pbResponse


class QueryResponseMessage:
    """
    Class for representing a query response message.

    Attributes:
        query_received (QueryMessageReceived): The received query message.
        client_id (str): The client ID.
        request_id (str): The request ID.
        is_executed (bool): Indicates if the query has been executed.
        timestamp (datetime): The timestamp of the query response.
        error (str): The error message, if any.
        metadata (str): The metadata associated with the query response.
        body (bytes): The body of the query response.
        tags (Dict[str, str]): The tags associated with the query response.

    Methods:
        validate() -> QueryResponseMessage:
            Validates the query response and raises an error if it is invalid.

        decode(pb_response: pbResponse) -> QueryResponseMessage:
            Decodes the protocol buffer response and updates the attributes of the query response message.

        encode(client_id: str) -> pbResponse:
            Encodes the query response message into a protocol buffer response.

        __repr__() -> str:
            Returns a string representation of the query response message.
    """
    def __init__(self, query_received: QueryMessageReceived = None,
                 metadata: str = None,
                 body: bytes = b'',
                 tags: Dict[str, str] = None,
                 is_executed: bool = False,
                 error: str = "",
                 timestamp: datetime = None,
                 ):
        self.query_received: QueryMessageReceived = query_received
        self.client_id: str = ""
        self.request_id: str = ""
        self.is_executed: bool = is_executed
        self.timestamp: datetime = timestamp if timestamp else datetime.now()
        self.error: str = error
        self.metadata: str = metadata
        self.body: bytes = body
        self.tags: Dict[str, str] = tags if tags else {}

    def validate(self) -> 'QueryResponseMessage':
        if not self.query_received:
            raise ValueError("Query response must have a query request.")
        elif self.query_received.reply_channel == "":
            raise ValueError("Query response must have a reply channel.")
        return self

    def decode(self, pb_response: pbResponse) -> 'QueryResponseMessage':
        self.client_id = pb_response.ClientID
        self.request_id = pb_response.RequestID
        self.is_executed = pb_response.Executed
        self.error = pb_response.Error
        self.timestamp = datetime.fromtimestamp(pb_response.Timestamp / 1e9)
        self.metadata = pb_response.Metadata
        self.body = pb_response.Body
        self.tags = pb_response.Tags
        return self

    def encode(self, client_id: str) -> pbResponse:
        pb_response = pbResponse()
        pb_response.ClientID = client_id
        pb_response.RequestID = self.query_received.id
        pb_response.ReplyChannel = self.query_received.reply_channel
        pb_response.Executed = self.is_executed
        pb_response.Error = self.error
        pb_response.Timestamp = int(self.timestamp.timestamp() * 1e9)
        pb_response.Metadata = self.query_received.metadata
        pb_response.Body = self.query_received.body
        for key, value in self.tags.items():
            pb_response.Tags[key] = value
        return pb_response

    def __repr__(self):
        return f"QueryResponseMessage: client_id={self.client_id}, request_id={self.request_id}, is_executed={self.is_executed}, error={self.error}, timestamp={self.timestamp}"



from typing import Callable
from kubemq.grpc import Subscribe
from kubemq.common.subscribe_type import SubscribeType
from kubemq.cq.query_message_received import QueryMessageReceived


class QueriesSubscription:
    """
    QueriesSubscription class represents a subscription to receive query messages from a channel.

    Attributes:
        channel (str): The name of the channel to subscribe to.
        group (str): The optional name of the group to subscribe to.
        on_receive_query_callback (Callable[[QueryMessageReceived], None]): The callback function to be called when a query message is received.
        on_error_callback (Callable[[str], None]): The callback function to be called when an error occurs.

    Methods:
        raise_on_receive_message(received_query: QueryMessageReceived) -> None:
            Raises the on_receive_query_callback with the received query message.

        raise_on_error(msg: str) -> None:
            Raises the on_error_callback with the specified error message.

        validate() -> None:
            Validates the query subscription by checking if the channel and on_receive_query_callback are set.

        encode(client_id: str = "") -> Subscribe:
            Encodes the query subscription into a Subscribe message.

        __repr__() -> str:
            Returns a string representation of the QueriesSubscription object.
    """
    def __init__(self, channel: str = None,
                 group: str = None,
                 on_receive_query_callback: Callable[[QueryMessageReceived], None] = None,
                 on_error_callback: Callable[[str], None] = None):
        self.channel: str = channel
        self.group: str = group
        self.on_receive_query_callback = on_receive_query_callback
        self.on_error_callback = on_error_callback


    def raise_on_receive_message(self, received_query: QueryMessageReceived):
        if self.on_receive_query_callback:
            self.on_receive_query_callback(received_query)

    def raise_on_error(self, msg: str):
        if self.on_error_callback:
            self.on_error_callback(msg)

    def validate(self):
        if not self.channel:
            raise ValueError("query subscription must have a channel.")
        if not self.on_receive_query_callback:
            raise ValueError("query subscription must have a on_receive_query_callback function.")

    def encode(self, client_id: str = "") -> Subscribe:
        request = Subscribe()
        request.Channel = self.channel
        request.Group = self.group
        request.ClientID = client_id
        request.SubscribeTypeData = SubscribeType.Queries.value
        return request

    def __repr__(self):
        return f"QueriesSubscription: channel={self.channel}, group={self.group}"


import os
from kubemq.queues import *


def list_queues_channels(search_pattern=""):
    try:
        with Client(address="localhost:50000", client_id="list_example") as client:
            list=client.list_queues_channels(search_pattern)
            print(f"Queues channels list: {list}")
    except Exception as e:
        print(f"Error listing Queues Channels: {e}")



if __name__ == "__main__":
    list_queues_channels()


import logging
import threading
import asyncio
import time
import uuid
import grpc
import socket
from typing import List
from kubemq.transport import *
from kubemq.queues.queues_message import *
from kubemq.queues import *
from kubemq.common.exceptions import *
from kubemq.common.helpers import *
from kubemq.common.requests import *
from kubemq.common.cancellation_token import *

class Client:
    """


    Client class represents a client that connects to the KubeMQ server.

    Attributes:
        connection (Connection): The connection configuration.
        logger (logging.Logger): The logger for logging events.
        transport (Transport): The transport layer used for communication.
        shutdown_event (threading.Event): The event used for notifying the client to shutdown.
        upstream_sender (UpstreamSender): The sender used for sending upstream messages.
        downstream_receiver (DownstreamReceiver): The receiver used for receiving downstream messages.

    Methods:
        connect(): Connects the client to the KubeMQ server.
        close(): Closes the connection to the KubeMQ server.
        ping() -> ServerInfo: Pings the KubeMQ server.
        send_queues_message(message: QueueMessage) -> QueueSendResult: Sends a message to the KubeMQ server queues.
        create_queues_channel(channel: str) -> [bool, None]: Creates a queues channel.
        delete_queues_channel(channel: str) -> [bool, None]: Deletes a queues channel.
        list_queues_channels(channel_search) -> List[QueuesChannel]: Lists the queues channels.
        receive_queues_messages(channel: str = None, max_messages: int = 1, wait_timeout_in_seconds: int = 60, auto_ack: bool = False) -> QueuesPollResponse: Receives messages from a queues
    * channel.

    Example usage:
        client = Client(address='localhost:50000')
        client.connect()
        server_info = client.ping()
        print(f"Server version: {server_info.server_version}")
        client.close()

    Note: Make sure to set the connection attributes correctly before connecting to the server.

    """
    def __init__(self, address: str = "",
                 client_id: str = socket.gethostname(),
                 auth_token: str = "",
                 tls: bool = False,
                 tls_cert_file: str = "",
                 tls_key_file: str = "",
                 tls_ca_file: str = "",
                 max_send_size: int = 0,
                 max_receive_size: int = 0,
                 disable_auto_reconnect: bool = False,
                 reconnect_interval_seconds: int = 0,
                 keep_alive: bool = False,
                 ping_interval_in_seconds: int = 0,
                 ping_timeout_in_seconds: int = 0,
                 log_level: int = None) -> None:
        self.connection: Connection = Connection(
            address=address,
            client_id=client_id,
            auth_token=auth_token,
            tls=TlsConfig(
                ca_file=tls_ca_file,
                cert_file=tls_cert_file,
                key_file=tls_key_file,
                enabled=tls,
            ),
            max_send_size=max_send_size,
            max_receive_size=max_receive_size,
            disable_auto_reconnect=disable_auto_reconnect,
            reconnect_interval_seconds=reconnect_interval_seconds,
            keep_alive=KeepAliveConfig(
                enabled=keep_alive,
                ping_interval_in_seconds=ping_interval_in_seconds,
                ping_timeout_in_seconds=ping_timeout_in_seconds
            ),
            log_level=log_level)
        self.logger = logging.getLogger("KubeMQ")
        if log_level is not None:
            self.logger.setLevel(log_level)
        else:
            self.logger.setLevel(logging.CRITICAL + 1)
        try:
            self.connection.validate()
            self.transport: Transport = Transport(self.connection).initialize()
            self.shutdown_event: threading.Event = threading.Event()
            self.upstream_sender = None
            self.downstream_receiver = None
        except ValueError as e:
            ex = ValidationError(e)
            self.logger.error(str(ex))
            raise ex
        except Exception as e:
            ex = GRPCError(e)
            self.logger.error(str(ex))
            raise ex

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def connect(self):
        try:
            self.logger.debug(f"Client connecting to {self.connection.address}")
            self.transport: Transport = Transport(self.connection).initialize()
            self.logger.debug(f"Client connected to {self.connection.address}")
        except Exception as e:
            ex =GRPCError(e)
            self.logger.error(str(ex))
            raise ex

    def close(self):
        try:
            self.shutdown_event.set()
            self.logger.debug(f"Client disconnecting from {self.connection.address}")
            self.transport.close()
        except Exception as e:
            ex = GRPCError(e)
            self.logger.error(str(ex))
            raise ex

    def ping(self) -> ServerInfo:
        """
        Ping the server and get the server information.

        Returns:
            ServerInfo: The server information.

        Raises:
            GRPCError: If an error occurs during the ping.

        """
        try:
            self.logger.debug(f"Client pinging {self.connection.address}")
            return self.transport.ping()
        except Exception as e:
            ex = GRPCError(e)
            self.logger.error(str(ex))
            raise ex

    def send_queues_message(self, message: QueueMessage) -> QueueSendResult:
        """
        Sends a message to the queues.

        Args:
            self: The current object instance.
            message (QueueMessage): The message to be sent.

        Returns:
            QueueSendResult: The result of the message send operation.
        """
        message.validate()
        if self.upstream_sender is None:
            self.upstream_sender = UpstreamSender(self.transport, self.shutdown_event, self.logger,self.connection)

        return self.upstream_sender.send(message)
    def create_queues_channel(self, channel: str)->[bool,None]:
        """
        Create a queues channel for the given channel.

        Parameters:
        - channel (str): The name of the channel to create.

        Returns:
        - bool or None: Returns True if the channel is created successfully, otherwise returns None.

        """
        return create_channel_request(self.transport, self.connection.client_id, channel, "queues")

    def delete_queues_channel(self, channel: str)->[bool,None]:
        """
        Delete Queues Channel

        Deletes a specific channel from the queues in the current connection.

        Parameters:
        - channel (str): The name of the channel to delete.

        Returns:
        - bool or None: Returns True if the channel was successfully deleted. Returns None if there was an error during deletion.
        """
        return delete_channel_request(self.transport, self.connection.client_id, channel, "queues")

    def list_queues_channels(self, channel_search)->List[QueuesChannel]:
        """
        Retrieve a list of QueuesChannel objects based on the provided channel_search.

        Parameters:
        :param channel_search: The search term used to filter the list of QueuesChannels.

        Returns:
        - A list of QueuesChannel objects that match the search term.
        """
        return list_queues_channels(self.transport, self.connection.client_id, channel_search)

    def receive_queues_messages(self, channel: str = None,
             max_messages: int = 1,
             wait_timeout_in_seconds: int = 60,
             auto_ack: bool = False,
             ) -> QueuesPollResponse:
        """

        Receive messages from a specific channel in the queues.

        Parameters:
        - channel (str): The name of the channel to receive messages from. Defaults to None.
        - max_messages (int): The maximum number of messages to receive. Defaults to 1.
        - wait_timeout_in_seconds (int): The maximum time in seconds to wait for new messages. Defaults to 60.
        - auto_ack (bool): Whether to automatically acknowledge the received messages. Defaults to False.

        Returns:
        - QueuesPollResponse: The response object containing the received messages.

        Raises:
        - ValueError: If the channel is None, max_messages is less than 1, or wait_timeout_in_seconds is less than 1.

        """
        if channel is None:
            raise ValueError("channel cannot be None.")
        if max_messages < 1:
            raise ValueError("poll_max_messages must be greater than 0.")
        if wait_timeout_in_seconds < 1:
            raise ValueError("poll_wait_timeout_in_seconds must be greater than 0.")

        if self.downstream_receiver is None:
            self.downstream_receiver = (
                DownstreamReceiver(self.transport,
                                         self.shutdown_event,
                                         self.logger,
                                         self.connection))

        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.connection.client_id
        request.Channel = channel
        request.MaxItems = max_messages
        request.WaitTimeout = wait_timeout_in_seconds * 1000
        request.AutoAck = auto_ack
        request.RequestTypeData = QueuesDownstreamRequestType.Get
        kubemq_response: QueuesDownstreamResponse = self.downstream_receiver.send(request)
        response = QueuesPollResponse().decode(
            response=kubemq_response,
            receiver_client_id=self.connection.client_id,
            response_handler=self.downstream_receiver.send_without_response)

        return response


import datetime
from typing import Callable
from kubemq.grpc import Subscribe
from kubemq.common.subscribe_type import SubscribeType
from kubemq.pubsub import EventStoreMessageReceived

from enum import Enum


class EventsStoreType(Enum):
    """
    Represents the type of event store.

    Enum Values:
    - Undefined: 0
    - StartNewOnly: 1
    - StartFromFirst: 2
    - StartFromLast: 3
    - StartAtSequence: 4
    - StartAtTime: 5
    - StartAtTimeDelta: 6
    """
    Undefined = 0
    StartNewOnly = 1
    StartFromFirst = 2
    StartFromLast = 3
    StartAtSequence = 4
    StartAtTime = 5
    StartAtTimeDelta = 6


class EventsStoreSubscription:
    """
    Class representing a subscription to an events store.

    Parameters:
    - channel (str): The channel to subscribe to.
    - group (str): The group to subscribe to.
    - events_store_type (EventsStoreType): The type of events store.
    - events_store_sequence_value (int): The sequence value of the events store.
    - events_store_start_time (datetime): The start time of the events store.
    - on_receive_event_callback (Callable[[EventStoreMessageReceived], None]): Callback function to handle received events.
    - on_error_callback (Callable[[str], None]): Callback function to handle errors.

    Methods:
    - raise_on_receive_message(received_event: EventStoreMessageReceived): Raises the on_receive_event_callback with the received event.
    - raise_on_error(msg: str): Raises the on_error_callback with the given error message.
    - validate(): Validates the subscription parameters.
    - encode(client_id: str = "") -> Subscribe: Encodes the subscription into a Subscribe request.
    - __repr__(): Returns a string representation of the subscription.

    """
    def __init__(self,
                 channel: str = None,
                 group: str = None,
                 events_store_type: EventsStoreType = EventsStoreType.Undefined,
                 events_store_sequence_value: int = 0,
                 events_store_start_time: datetime = None,
                 on_receive_event_callback: Callable[[EventStoreMessageReceived], None] = None,
                 on_error_callback: Callable[[str], None] = None):
        self.channel: str = channel
        self.group: str = group
        self.events_store_type: EventsStoreType = events_store_type
        self.events_store_sequence_value: int = events_store_sequence_value
        self.events_store_start_time: datetime = events_store_start_time
        self.on_receive_event_callback = on_receive_event_callback
        self.on_error_callback = on_error_callback

    def raise_on_receive_message(self, received_event: EventStoreMessageReceived):
        if self.on_receive_event_callback:
            self.on_receive_event_callback(received_event)

    def raise_on_error(self, msg: str):
        if self.on_error_callback:
            self.on_error_callback(msg)

    def validate(self):
        if not self.channel:
            raise ValueError("Event Store subscription must have a channel.")
        if not self.on_receive_event_callback:
            raise ValueError("Event Store subscription must have a on_receive_event_callback function.")
        if self.events_store_type == EventsStoreType.Undefined:
            raise ValueError("Event Store subscription must have an events store type.")
        if self.events_store_type == EventsStoreType.StartAtSequence and self.events_store_sequence_value == 0:
            raise ValueError("Event Store subscription with StartAtSequence events store type must have a sequence value.")
        if self.events_store_type == EventsStoreType.StartAtTime and not self.events_store_start_time:
            raise ValueError("Event Store subscription with StartAtTime events store type must have a start time.")

    def encode(self, client_id: str = "") -> Subscribe:
        request = Subscribe()
        request.Channel = self.channel
        request.Group = self.group
        if self.events_store_type == EventsStoreType.StartNewOnly:
            request.EventsStoreTypeData = EventsStoreType.StartNewOnly.value
        elif self.events_store_type == EventsStoreType.StartFromFirst:
            request.EventsStoreTypeData = EventsStoreType.StartFromFirst.value
        elif self.events_store_type == EventsStoreType.StartFromLast:
            request.EventsStoreTypeData = EventsStoreType.StartFromLast.value
        elif self.events_store_type == EventsStoreType.StartAtSequence:
            request.EventsStoreTypeData = self.events_store_sequence_value
            request.EventsStoreTypeValue = self.events_store_sequence_value
        elif self.events_store_type == EventsStoreType.StartAtTime:
            request.EventsStoreTypeData = EventsStoreType.StartAtTime.value
            request.EventsStoreTypeValue = int(self.events_store_start_time.timestamp())
        request.ClientID = client_id
        request.SubscribeTypeData = SubscribeType.EventsStore.value
        return request

    def __repr__(self):
        return f"EventsStoreSubscription: channel={self.channel}, group={self.group}, events_store_type={self.events_store_type.name}, events_store_sequence_value={self.events_store_sequence_value}, events_store_start_time={self.events_store_start_time}"


import uuid
from typing import Dict, Optional
from kubemq.grpc import Request as pbQuery


class QueryMessage:
    """
    Class representing a query message.

    Attributes:
        id (str): The ID of the query message.
        channel (str): The channel of the query message.
        metadata (str): The metadata of the query message.
        body (bytes): The body of the query message.
        tags (Dict[str, str]): The tags of the query message.
        timeout_in_seconds (int): The timeout of the query message in seconds.
        cache_key (str): The cache key of the query message.
        cache_ttl_int_seconds (int): The cache TTL of the query message in seconds.

    Methods:
        validate(): Validates the query message.
        encode(client_id: str) -> pbQuery: Encodes the query message into a protobuf query.
        __repr__(): Returns a string representation of the query message.
    """
    def __init__(self, id: str = None,
                 channel: str = None,
                 metadata: str = None,
                 body: bytes = b'',
                 tags: Dict[str, str] = None,
                 timeout_in_seconds: int = 0,
                 cache_key: str = "",
                 cache_ttl_int_seconds: int = 0,
                 ):
        self.id: str = id
        self.channel: str = channel
        self.metadata: str = metadata
        self.body: bytes = body
        self.tags: Dict[str, str] = tags if tags else {}
        self.timeout_in_seconds: int = timeout_in_seconds
        self.cache_key: str = cache_key
        self.cache_ttl_int_seconds: int = cache_ttl_int_seconds

    def validate(self) -> 'QueryMessage':
        if not self.channel:
            raise ValueError("Query message must have a channel.")

        if not self.metadata and not self.body and not self.tags:
            raise ValueError("Query message must have at least one of the following: metadata, body, or tags.")

        if self.timeout_in_seconds <= 0:
            raise ValueError("Query message timeout must be a positive integer.")
        return self

    def encode(self, client_id: str) -> pbQuery:

        pb_query = pbQuery()  # Assuming pb is an imported module
        pb_query.RequestID = self.id or str(uuid.uuid4())
        pb_query.ClientID = client_id
        pb_query.Channel = self.channel
        pb_query.Metadata = self.metadata or ""
        pb_query.Body = self.body
        pb_query.Timeout = self.timeout_in_seconds * 1000
        pb_query.RequestTypeData = pbQuery.RequestType.Query
        for key, value in self.tags.items():
            pb_query.Tags[key] = value
        pb_query.CacheKey = self.cache_key
        pb_query.CacheTTL = self.cache_ttl_int_seconds * 1000
        return pb_query

    def __repr__(self):
        return f"QueryMessage: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, tags={self.tags}, timeout_in_seconds={self.timeout_in_seconds}, cache_key={self.cache_key}, cache_ttl_int_seconds={self.cache_ttl_int_seconds}"


import grpc
def decode_grpc_error(exc) -> str:
    """
    Decodes a gRPC error and returns a formatted error message.

    Parameters:
    - exc: The gRPC exception to decode.

    Returns:
    - str: The formatted error message.

    """
    message = str(exc)

    # Check if the exception has 'code' (status) and 'details' methods
    if hasattr(exc, 'code') and callable(exc.code) and hasattr(exc, 'details') and callable(exc.details):
        status = exc.code()
        details = exc.details()
        if status == grpc.StatusCode.CANCELLED:
            return "KubeMQ Connection Closed"
        # Ensure that status and details are not None
        if status is not None and details is not None:
            message = f"KubeMQ Connection Error - Status: {status} Details: {details}"

    return message


from datetime import datetime, timedelta
from typing import Dict
from kubemq.grpc import Request as pbRequest


class QueryMessageReceived:
    """
    Class representing a query message received.

    Attributes:
        id (str): The unique ID of the query message.
        from_client_id (str): The ID of the client that sent the query message.
        timestamp (datetime): The timestamp when the query message was received.
        channel (str): The channel through which the query message was received.
        metadata (str): Additional metadata associated with the query message.
        body (bytes): The body of the query message.
        reply_channel (str): The channel through which the reply for the query message should be sent.
        tags (dict): A dictionary containing tags associated with the query message.

    Methods:
        decode(query_receive: pbRequest) -> QueryMessageReceived:
            Decodes a protobuf request object and returns a QueryMessageReceived instance.

        __repr__() -> str:
            Returns a string representation of the QueryMessageReceived object.
    """
    def __init__(self):
        self.id: str = ""
        self.from_client_id: str = ""
        self.timestamp: datetime = datetime.fromtimestamp(0)
        self.channel: str = ""
        self.metadata: str = ""
        self.body: bytes = b""
        self.reply_channel: str = ""
        self.tags: Dict[str, str] = {}

    @staticmethod
    def decode(query_receive: pbRequest) -> 'QueryMessageReceived':
        message = QueryMessageReceived()
        message.id = query_receive.RequestID
        message.from_client_id = query_receive.ClientID
        message.timestamp = datetime.now()
        message.channel = query_receive.Channel
        message.metadata = query_receive.Metadata
        message.body = query_receive.Body
        message.reply_channel = query_receive.ReplyChannel
        message.tags = query_receive.Tags
        return message

    def __repr__(self):
        return f"QueryMessageReceived: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, from_client_id={self.from_client_id}, timestamp={self.timestamp}, reply_channel={self.reply_channel}, tags={self.tags}"



from typing import Callable
from kubemq.grpc import Subscribe
from kubemq.common.subscribe_type import SubscribeType
from kubemq.cq.command_message_received import CommandMessageReceived


class CommandsSubscription:
    """
    Class representing a subscription to receive commands.

    Attributes:
        channel (str): The channel to subscribe to.
        group (str): The group to subscribe to.
        on_receive_command_callback (Callable[[CommandMessageReceived], None]): Callback function to handle received commands.
        on_error_callback (Callable[[str], None]): Callback function to handle errors.

    Methods:
        raise_on_receive_message(received_command: CommandMessageReceived): Raises the on_receive_command_callback function with the received command.
        raise_on_error(msg: str): Raises the on_error_callback function with the error message.
        validate(): Validates the command subscription by checking if channel and on_receive_command_callback are set.
        decode(client_id: str = "") -> Subscribe: Decodes the subscription into a Subscribe object.
        __repr__(): Returns a string representation of the CommandsSubscription object.
    """
    def __init__(self, channel: str = None,
                 group: str = None,
                 on_receive_command_callback: Callable[[CommandMessageReceived], None] = None,
                 on_error_callback: Callable[[str], None] = None):
        self.channel: str = channel
        self.group: str = group
        self.on_receive_command_callback = on_receive_command_callback
        self.on_error_callback = on_error_callback

    def raise_on_receive_message(self, received_command: CommandMessageReceived):
        if self.on_receive_command_callback:
            self.on_receive_command_callback(received_command)

    def raise_on_error(self, msg: str):
        if self.on_error_callback:
            self.on_error_callback(msg)

    def validate(self):
        if not self.channel:
            raise ValueError("command subscription must have a channel.")
        if not self.on_receive_command_callback:
            raise ValueError("command subscription must have a on_receive_command_callback function.")

    def decode(self, client_id: str = "") -> Subscribe:
        request = Subscribe()
        request.Channel = self.channel
        request.Group = self.group
        request.ClientID = client_id
        request.SubscribeTypeData = SubscribeType.Commands.value
        return request

    def __repr__(self):
        return f"CommandsSubscription: channel={self.channel}, group={self.group}"


import uuid
from typing import Callable
from kubemq.grpc import QueuesDownstreamRequest, QueuesDownstreamResponse, QueuesDownstreamRequestType
from kubemq.queues.queues_message_received import QueueMessageReceived


class QueuesPollResponse:
    """
    This class represents the response received when polling a queue.

    Attributes:
        ref_request_id (str): The reference request ID.
        transaction_id (str): The transaction ID.
        messages (List[QueueMessageReceived]): The list of queue messages received.
        error (str): The error message, if any.
        is_error (bool): A flag indicating whether an error occurred.
        is_transaction_completed (bool): A flag indicating whether the transaction is completed.
        active_offsets (List[int]): The active offsets.
        response_handler (Callable[[QueuesDownstreamRequest], None]): The response handler function.
        receiver_client_id (str): The receiver client ID.

    Methods:
        ack_all: Sends an acknowledgement for all messages in the transaction.
        reject_all: Rejects all messages in the transaction.
        re_queue_all: Re-queues all messages in the transaction to a specified channel.
        decode: Decodes the response received from the downstream and populates the response object.

    """
    def __init__(self):
        self.ref_request_id: str = ""
        self.transaction_id: str = ""
        self.messages: [QueueMessageReceived] = []
        self.error: str = ""
        self.is_error: bool = False
        self.is_transaction_completed: bool = False
        self.active_offsets: [int] = []
        self.response_handler: Callable[[QueuesDownstreamRequest], None]
        self.receiver_client_id: str = ""

    def ack_all(self):
        if self.is_transaction_completed:
            raise ValueError("transaction is already completed")
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.receiver_client_id
        request.RequestTypeData = QueuesDownstreamRequestType.AckAll
        request.RefTransactionId = self.transaction_id
        request.SequenceRange.extend(self.active_offsets)
        self.response_handler(request)

    def reject_all(self):
        if self.is_transaction_completed:
            raise ValueError("transaction is already completed")
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.receiver_client_id
        request.RequestTypeData = QueuesDownstreamRequestType.NAckAll
        request.RefTransactionId = self.transaction_id
        request.SequenceRange.extend(self.active_offsets)
        result = self.response_handler(request)
        if result.IsError:
            raise ValueError(result.Error)

    def re_queue_all(self, channel: str):
        if not channel:
            raise ValueError("re-queue channel cannot be empty")
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.receiver_client_id
        request.RequestTypeData = QueuesDownstreamRequestType.ReQueueAll
        request.RefTransactionId = self.transaction_id
        request.SequenceRange.extend(self.active_offsets)
        request.ReQueueChannel = channel
        result = self.response_handler(request)
        if result.IsError:
            raise ValueError(result.Error)

    def decode(self,
               response: QueuesDownstreamResponse,
               receiver_client_id: str,
               response_handler: Callable[
                                      [QueuesDownstreamRequest], QueuesDownstreamResponse]) -> 'QueuesPollResponse':
        self.ref_request_id = response.RefRequestId
        self.transaction_id = response.TransactionId
        self.error = response.Error
        self.is_error = response.IsError
        self.is_transaction_completed = response.TransactionComplete
        self.active_offsets.extend([response.ActiveOffsets])
        self.response_handler = response_handler
        self.receiver_client_id = receiver_client_id
        for message in response.Messages:
            self.messages.append(QueueMessageReceived().decode(
                message,
                response.TransactionId,
                response.TransactionComplete,
                receiver_client_id,
                response_handler))
        return self



import logging
import time
from kubemq.pubsub import *

def main():
    try:
        with Client(address="localhost:50000", client_id="events_store_example") as client:
            def on_receive_event(event: EventStoreMessageReceived):
                print(
                    f"Id:{event.id}, Timestamp:{event.timestamp} From: {event.from_client_id},  Body:{event.body.decode('utf-8')}")
            def on_error_handler(err:str):
                print(f"{err}")

            client.subscribe_to_events_store(
                subscription=EventsStoreSubscription(
                    channel="es1",
                    group="",
                    on_receive_event_callback=on_receive_event,
                    on_error_callback=on_error_handler,
                    events_store_type=EventsStoreType.StartNewOnly,
                )
                , cancel= CancellationToken())

            time.sleep(1)
            result= client.send_events_store_message(EventStoreMessage(
                channel="es1",
                body=b"hello kubemq"
            ))
            print(f"send result:{result}")
            time.sleep(1)
    except Exception as e:
        print(e)
        return

if __name__ == "__main__":
    main()



import uuid
from typing import Dict, Optional
from kubemq.grpc import Event as pbEvent


class EventStoreMessage:
    """
    EventStoreMessage

    Class representing a message to be stored in an event store.

    Attributes:
        id (str): Unique identifier for the message. If not provided, a random UUID will be generated.
        channel (str): Channel for the message.
        metadata (str): Metadata associated with the message.
        body (bytes): Body of the message.
        tags (Dict[str, str]): Additional tags associated with the message.

    Methods:
        validate() -> 'EventStoreMessage':
            Validates that the necessary fields are present in the message.
            Raises a ValueError if any required field is missing.
            Returns self.

        encode(client_id: str) -> pbEvent:
            Encodes the message into a pbEvent object for transmission.
            Adds the client_id as a tag to the message.
            Returns the pbEvent object.

        __repr__() -> str:
            Returns a string representation of the EventStoreMessage object.
    """
    def __init__(self, id: str = None,
                 channel: str = None,
                 metadata: str = None,
                 body: bytes = b'',
                 tags: Dict[str, str] = None):
        self.id: str = id
        self.channel: str = channel
        self.metadata: str = metadata
        self.body: bytes = body
        self.tags: Dict[str, str] = tags if tags else {}

    def validate(self) -> 'EventStoreMessage':
        if not self.channel:
            raise ValueError("Event Store message must have a channel.")

        if not self.metadata and not self.body and not self.tags:
            raise ValueError("Event Store message must have at least one of the following: metadata, body, or tags.")

        return self

    def encode(self, client_id: str) -> pbEvent:
        self.tags["x-kubemq-client-id"] = client_id
        pb_event = pbEvent()
        pb_event.EventID = self.id or str(uuid.uuid4())
        pb_event.ClientID = client_id
        pb_event.Channel = self.channel
        pb_event.Metadata = self.metadata or ""
        pb_event.Body = self.body
        pb_event.Store = True
        for key, value in self.tags.items():
            pb_event.Tags[key] = value
        return pb_event

    def __repr__(self):
        return f"EventStoreMessage: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, tags={self.tags}"


from datetime import datetime
from kubemq.cq.command_message_received import CommandMessageReceived
from kubemq.grpc import Response as pbResponse


class CommandResponseMessage:
    """

    Class representing a command response message.

    Attributes:
        command_received (CommandMessageReceived): The received command message this response is associated with.
        client_id (str): The client ID associated with this response.
        request_id (str): The request ID associated with this response.
        is_executed (bool): Indicates if the command associated with this response was executed successfully.
        timestamp (datetime): The timestamp of this response.
        error (str): The error message associated with this response.

    Methods:
        __init__(command_received=None, is_executed=False, error="", timestamp=None):
            Initializes a new CommandResponseMessage object with the given parameters.

        validate() -> CommandResponseMessage:
            Validates the command response message. Throws a ValueError if the command response is invalid.

        decode(pb_response) -> CommandResponseMessage:
            Decodes the protocol buffer response and populates the CommandResponseMessage attributes.

        encode(client_id) -> pbResponse:
            Encodes the CommandResponseMessage into a protocol buffer response object.

        __repr__() -> str:
            Returns a string representation of the CommandResponseMessage.

    """
    def __init__(self, command_received: CommandMessageReceived = None,
                 is_executed: bool = False,
                 error: str = "",
                 timestamp: datetime = None,
                 ):
        self.command_received: CommandMessageReceived = command_received
        self.client_id: str = ""
        self.request_id: str = ""
        self.is_executed: bool = is_executed
        self.timestamp: datetime = timestamp if timestamp else datetime.now()
        self.error: str = error
    def validate(self) -> 'CommandResponseMessage':
        if not self.command_received:
            raise ValueError("Command response must have a command request.")
        elif self.command_received.reply_channel == "":
            raise ValueError("Command response must have a reply channel.")
        return self

    def decode(self, pb_response: pbResponse) -> 'CommandResponseMessage':
        self.client_id = pb_response.ClientID
        self.request_id = pb_response.RequestID
        self.is_executed = pb_response.Executed
        self.error = pb_response.Error
        self.timestamp = datetime.fromtimestamp(pb_response.Timestamp / 1e9)
        return self

    def encode(self, client_id: str) -> pbResponse:
        pb_response = pbResponse()
        pb_response.ClientID = client_id
        pb_response.RequestID = self.command_received.id
        pb_response.ReplyChannel = self.command_received.reply_channel
        pb_response.Executed = self.is_executed
        pb_response.Error = self.error
        pb_response.Timestamp = int(self.timestamp.timestamp() * 1e9)
        return pb_response

    def __repr__(self):
        return f"CommandResponseMessage: client_id={self.client_id}, request_id={self.request_id}, is_executed={self.is_executed}, error={self.error}, timestamp={self.timestamp}"


import grpc
from grpc import Channel
from grpc._cython.cygrpc import ChannelCredentials
import kubemq.grpc.kubemq_pb2_grpc as kubemq_pb2_grpc
from typing import Sequence
from kubemq.transport.connection import Connection
from kubemq.transport.tls_config import TlsConfig
from kubemq.transport.keep_alive import KeepAliveConfig
from kubemq.transport._interceptors import _AuthInterceptors, _AuthInterceptorsAsync
from kubemq.transport.server_info import ServerInfo
from kubemq.grpc import Empty


def _get_ssl_credentials(tls_config: TlsConfig) -> ChannelCredentials:
    certificate_chain = _read_file(tls_config.cert_file) if tls_config.cert_file else None
    private_key = _read_file(tls_config.key_file) if tls_config.key_file else None
    root_certificates = _read_file(tls_config.ca_file) if tls_config.ca_file else None
    return grpc.ssl_channel_credentials(root_certificates, private_key, certificate_chain)


def _read_file(file_path):
    with open(file_path, 'rb') as f:
        return f.read()


def _get_call_options(connection: Connection) -> Sequence:
    options = [
        ('grpc.max_send_message_length', connection.max_send_size),
        ('grpc.max_receive_message_length', connection.max_receive_size),
    ]
    if connection.keep_alive and isinstance(connection.keep_alive, KeepAliveConfig) and connection.keep_alive.enabled:
        options.append(('grpc.keepalive_time_ms', connection.keep_alive.ping_timeout_in_seconds * 1000))
        options.append(('grpc.keepalive_timeout_ms', connection.keep_alive.ping_interval_in_seconds * 1000))
        options.append(('grpc.keepalive_permit_without_calls', 1))
        options.append(('grpc.http2.min_time_between_pings_ms', connection.keep_alive.ping_timeout_in_seconds * 1000))
        options.append(
            ('grpc.http2.min_ping_interval_without_data_ms', connection.keep_alive.ping_interval_in_seconds * 1000))
    return options


class Transport:
    def __init__(self, connection: Connection) -> None:
        self._opts: Connection = connection.complete()
        self._channel: Channel = None
        self._client: kubemq_pb2_grpc.kubemqStub = None
        self._async_channel: Channel = None
        self._async_client: kubemq_pb2_grpc.kubemqStub = None
        self._is_connected: bool = False

    def initialize(self) -> 'Transport':
        try:
            auth_interceptor = _AuthInterceptors(self._opts.auth_token)
            interceptors = [auth_interceptor]
            credentials = _get_ssl_credentials(self._opts.tls) if self._opts.tls.enabled else None
            self._channel = grpc.secure_channel(self._opts.address, credentials, options=_get_call_options(
                self._opts)) if credentials else grpc.insecure_channel(self._opts.address,
                                                                       options=_get_call_options(self._opts))
            self._channel = grpc.intercept_channel(self._channel, *interceptors)
            self._client = kubemq_pb2_grpc.kubemqStub(self._channel)
            self.ping()
            self._initialize_async()
            self._is_connected = True
        except Exception as ex:
            self._is_connected = False
            raise ex
        return self

    def _initialize_async(self) -> None:
        auth_interceptor_async: _AuthInterceptorsAsync = _AuthInterceptorsAsync(self._opts.auth_token)
        interceptors_async: Sequence[grpc.aio.ClientInterceptor] = [auth_interceptor_async]
        if self._opts.tls.enabled:
            try:
                credentials: ChannelCredentials = _get_ssl_credentials(self._opts.tls)
                self._async_channel = grpc.aio.secure_channel(self._opts.address, credentials,
                                                              options=_get_call_options(self._opts),
                                                              interceptors=interceptors_async)
            except Exception as e:
                raise e
        else:
            self._async_channel = grpc.aio.insecure_channel(self._opts.address, options=_get_call_options(self._opts),
                                                            interceptors=interceptors_async)

        self._async_client = kubemq_pb2_grpc.kubemqStub(self._channel)

    def ping(self) -> ServerInfo:
        response = self._client.Ping(Empty())
        return ServerInfo(
            host=response.Host,
            version=response.Version,
            server_start_time=response.ServerStartTime,
            server_up_time_seconds=response.ServerUpTimeSeconds
        )

    def kubemq_client(self) -> kubemq_pb2_grpc.kubemqStub:
        return self._client

    def kubemq_async_client(self) -> kubemq_pb2_grpc.kubemqStub:
        return self._async_client

    def is_connected(self) -> bool:
        return self._is_connected

    def close(self) -> None:
        if self._is_connected and self._channel is not None:
            self._channel.close()
            self._is_connected = False
            self._channel = None
            self._client = None


class AsyncTransport:
    def __init__(self, connection: Connection) -> None:
        self._opts: Connection = connection.complete()
        self._channel: Channel = None
        self._client: kubemq_pb2_grpc.kubemqStub = None
        self._is_connected: bool = False

    async def initialize(self) -> 'AsyncTransport':
        auth_interceptor_async: _AuthInterceptorsAsync = _AuthInterceptorsAsync(self._opts.auth_token)
        interceptors_async: Sequence[grpc.aio.ClientInterceptor] = [auth_interceptor_async]

        if self._opts.tls.enabled:
            try:
                credentials: ChannelCredentials = _get_ssl_credentials(self._opts.tls)
                self._channel = grpc.aio.secure_channel(self._opts.address, credentials,
                                                        options=_get_call_options(self._opts),
                                                        interceptors=interceptors_async)
            except Exception as e:
                raise e
        else:
            self._channel = grpc.aio.insecure_channel(self._opts.address, options=_get_call_options(self._opts),
                                                      interceptors=interceptors_async)

        self._client = kubemq_pb2_grpc.kubemqStub(self._channel)
        try:
            await self.ping()
            self._is_connected = True
        except Exception as ex:
            self._is_connected = False
            raise ex
        return self

    async def ping(self) -> ServerInfo:
        response = await self._client.Ping(Empty())
        return ServerInfo(
            host=response.Host,
            version=response.Version,
            server_start_time=response.ServerStartTime,
            server_up_time_seconds=response.ServerUpTimeSeconds
        )

    def kubemq_client(self) -> kubemq_pb2_grpc.kubemqStub:
        return self._client

    def is_connected(self) -> bool:
        return self._is_connected

    async def close(self) -> None:
        if self._is_connected and self._channel is not None:
            await self._channel.close()
            self._is_connected = False
            self._channel = None
            self._client = None



import datetime
from typing import Dict
from kubemq.grpc import Request as pbRequest


class CommandMessageReceived:
    """
    CommandMessageReceived class represents a received command message.

    Attributes:
        id (str): The ID of the command message.
        from_client_id (str): The ID of the client that sent the command message.
        timestamp (datetime): The timestamp when the command message was received.
        channel (str): The channel on which the command message was received.
        metadata (str): Additional metadata associated with the command message.
        body (bytes): The body of the command message.
        reply_channel (str): The channel to which any reply to the command message should be sent.
        tags (Dict[str, str]): Additional tags associated with the command message.

    Methods:
        decode(command_receive: pbRequest) -> CommandMessageReceived:
            Static method that takes a pbRequest object as input and decodes it into a CommandMessageReceived object.

        __repr__() -> str:
            Returns a string representation of the CommandMessageReceived object.

    """
    def __init__(self):
        self.id: str = ""
        self.from_client_id: str = ""
        self.timestamp: datetime = datetime.datetime.fromtimestamp(0)
        self.channel: str = ""
        self.metadata: str = ""
        self.body: bytes = b""
        self.reply_channel: str = ""
        self.tags: Dict[str, str] = {}

    @staticmethod
    def decode(command_receive: pbRequest) -> 'CommandMessageReceived':
        message = CommandMessageReceived()
        message.id = command_receive.RequestID
        message.from_client_id = command_receive.ClientID
        message.timestamp = datetime.datetime.now()
        message.channel = command_receive.Channel
        message.metadata = command_receive.Metadata
        message.body = command_receive.Body
        message.reply_channel = command_receive.ReplyChannel
        return message

    def __repr__(self):
        return f"CommandMessageReceived: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, from_client_id={self.from_client_id}, timestamp={self.timestamp}, reply_channel={self.reply_channel}, tags={self.tags}"



import datetime
from kubemq.grpc import SendQueueMessageResult


class QueueSendResult:
    """

    QueueSendResult represents the result of sending a message to a queue.

    Attributes:
        id (str): The unique identifier of the message.
        sent_at (datetime): The timestamp when the message was sent.
        expired_at (datetime): The timestamp when the message will expire.
        delayed_to (datetime): The timestamp when the message will be delivered.
        is_error (bool): Indicates if there was an error while sending the message.
        error (str): The error message if `is_error` is True.

    Methods:
        decode(result: SendQueueMessageResult) -> QueueSendResult:
            Decodes the given SendQueueMessageResult and sets the attributes of the QueueSendResult instance accordingly.
            Returns the updated QueueSendResult instance.

        __repr__() -> str:
            Returns a string representation of the QueueSendResult object.

    """
    def __init__(self, id: str = None,
                 sent_at: datetime = None,
                 expired_at: datetime = None,
                 delayed_to: datetime = None,
                 is_error: bool = False,
                 error: str = None):
        self.id: str = id
        self.sent_at: datetime = sent_at
        self.expired_at: datetime = expired_at
        self.delayed_to: datetime = delayed_to
        self.is_error: bool = is_error
        self.error: str = error

    def decode(self, result:SendQueueMessageResult) -> 'QueueSendResult':
        self.id = result.MessageID if result.MessageID else ""
        self.sent_at = datetime.datetime.fromtimestamp(result.SentAt / 1e9) if result.SentAt > 0 else None
        self.expired_at = datetime.datetime.fromtimestamp(result.ExpirationAt /1e9) if result.ExpirationAt > 0 else None
        self.delayed_to = datetime.datetime.fromtimestamp(result.DelayedTo /1e9) if result.DelayedTo > 0 else None
        self.is_error = result.IsError if result.IsError else False
        self.error = result.Error if result.Error else ""
        return self

    def __repr__(self):
        return f"QueueSendResult: id={self.id}, sent_at={self.sent_at}, expired_at={self.expired_at}, delayed_to={self.delayed_to}, is_error={self.is_error}, error={self.error}"


from datetime import datetime
from typing import Dict
from kubemq.grpc import EventReceive as pbEventReceive


class EventMessageReceived:
    """

    This class represents a received event message.

    Attributes:
        id (str): The unique identifier of the event message.
        from_client_id (str): The client ID from which the event message originated.
        timestamp (datetime): The timestamp when the event message was received.
        channel (str): The channel to which the event message was sent.
        metadata (str): The metadata associated with the event message.
        body (bytes): The body of the event message.
        tags (Dict[str, str]): Additional tags associated with the event message.

    Methods:
        decode(event_receive: pbEventReceive) -> EventMessageReceived:
            Static method that decodes an event_receive object of type pbEventReceive and returns an instance of EventMessageReceived.

        __repr__() -> str:
            Returns a string representation of the EventMessageReceived object.

    """
    def __init__(self):
        self.id: str = ""
        self.from_client_id: str = ""
        self.timestamp: datetime = datetime.now()
        self.channel: str = ""
        self.metadata: str = ""
        self.body: bytes = b""
        self.tags: Dict[str, str] = {}

    @staticmethod
    def decode(event_receive: pbEventReceive) -> 'EventMessageReceived':
        from_client_id = event_receive.Tags.get("x-kubemq-client-id", "") if event_receive.Tags else ""
        tags = event_receive.Tags if event_receive.Tags else {}
        message = EventMessageReceived()
        message.id = event_receive.EventID
        message.from_client_id = from_client_id
        message.channel = event_receive.Channel
        message.metadata = event_receive.Metadata
        message.body = event_receive.Body
        message.tags = tags
        return message

    def __repr__(self):
        return f"EventMessageReceived: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, from_client_id={self.from_client_id}, timestamp={self.timestamp}, tags={self.tags}"


import uuid
from typing import Dict
from kubemq.grpc import Request as pbCommand


class CommandMessage:
    """

    The `CommandMessage` class represents a command message that can be sent over a communication channel. It contains information such as the message ID, channel, metadata, body, tags,
    * and timeout.

    Attributes:
    - `id` (str): The ID of the command message.
    - `channel` (str): The channel through which the command message will be sent.
    - `metadata` (str): Additional metadata associated with the command message.
    - `body` (bytes): The body of the command message as bytes.
    - `tags` (dict): A dictionary of key-value pairs representing tags associated with the command message.
    - `timeout_in_seconds` (int): The maximum time in seconds for which the command message is valid.

    Methods:
    - `validate() -> CommandMessage`: Validates the command message. Raises a `ValueError` if the message is invalid.
    - `encode(client_id: str) -> pbCommand`: Encodes the command message into a protocol buffer message. Returns the encoded message.
    - `__repr__() -> str`: Returns a string representation of the command message.

    Example usage:
    command = CommandMessage(id='123', channel='commands', metadata='Execute command', body=b'command body', tags={'tag1': 'value1'}, timeout_in_seconds=10)
    command.validate()
    encoded_message = command.encode('client1')
    print(encoded_message)

    """
    def __init__(self, id: str = None,
                 channel: str = None,
                 metadata: str = None,
                 body: bytes = b'',
                 tags: Dict[str, str] = None,
                 timeout_in_seconds: int = 0):
        self.id: str = id
        self.channel: str = channel
        self.metadata: str = metadata
        self.body: bytes = body
        self.tags: Dict[str, str] = tags if tags else {}
        self.timeout_in_seconds: int = timeout_in_seconds

    def validate(self) -> 'CommandMessage':
        if not self.channel:
            raise ValueError("Command message must have a channel.")

        if not self.metadata and not self.body and not self.tags:
            raise ValueError("Command message must have at least one of the following: metadata, body, or tags.")

        if self.timeout_in_seconds <= 0:
            raise ValueError("Command message timeout must be a positive integer.")
        return self

    def encode(self, client_id: str) -> pbCommand:
        pb_command = pbCommand()
        pb_command.RequestID = self.id or str(uuid.uuid4())
        pb_command.ClientID = client_id
        pb_command.Channel = self.channel
        pb_command.Metadata = self.metadata or ""
        pb_command.Body = self.body
        pb_command.Timeout = self.timeout_in_seconds * 1000
        pb_command.RequestTypeData = pbCommand.RequestType.Command
        for key, value in self.tags.items():
            pb_command.Tags[key] = value
        return pb_command

    def __repr__(self):
        return f"CommandMessage: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, tags={self.tags}, timeout_in_seconds={self.timeout_in_seconds}"


import logging
from kubemq.queues import *

def main():
    with Client(address="localhost:50000") as client:
        send_result= client.send_queues_message(
                QueueMessage(
                    channel="q1",
                    body=b"some-simple_queue-queue-message",
                    metadata="some-metadata",
                )
            )
        print(f"Queue Message Sent: {send_result}")

        result = client.receive_queues_messages(
            channel="q1",
            max_messages=1,
            wait_timeout_in_seconds=10,
            auto_ack=False,
        )
        if result.is_error:
            print(f"{result.error}")
            return
        for message in result.messages:
            print(f"Id:{message.id}, Body:{message.body.decode('utf-8')}")
            message.ack()



if __name__ == "__main__":
    main()



from datetime import datetime, timedelta
from typing import Dict, ByteString
from kubemq.grpc import EventReceive as pbEventReceive


class EventStoreMessageReceived:
    """
    Class EventStoreMessageReceived

    Represents a received message in the EventStore.

    Properties:
    - id: str - The ID of the message.
    - from_client_id: str - The ID of the client that sent the message.
    - timestamp: datetime - The timestamp of when the message was received.
    - channel: str - The channel to which the message was published.
    - metadata: str - The metadata associated with the message.
    - body: bytes - The body of the message.
    - sequence: int - The sequence number of the message.
    - tags: Dict[str, str] - The tags associated with the message.

    Methods:
    - decode(event_receive: pbEventReceive) -> 'EventStoreMessageReceived':
      Static method that decodes a protobuf event_receive object and returns an instance
      of EventStoreMessageReceived with the decoded values.

    - __repr__() -> str:
      Returns a string representation of the EventStoreMessageReceived object.

    Example usage:
    ```
    message = EventStoreMessageReceived.decode(event_receive)
    print(message)
    ```
    """
    def __init__(self):
        self.id: str = ""
        self.from_client_id: str = ""
        self.timestamp: datetime = datetime.fromtimestamp(0)
        self.channel: str = ""
        self.metadata: str = ""
        self.body: bytes = b""
        self.sequence: int = 0
        self.tags: Dict[str, str] = {}

    @staticmethod
    def decode(event_receive: pbEventReceive) -> 'EventStoreMessageReceived':
        from_client_id = event_receive.Tags.get("x-kubemq-client-id", "") if event_receive.Tags else ""
        tags = event_receive.Tags if event_receive.Tags else {}
        message = EventStoreMessageReceived()
        message.id = event_receive.EventID
        message.from_client_id = from_client_id
        message.timestamp = datetime.fromtimestamp(event_receive.Timestamp / 1e9)
        message.channel = event_receive.Channel
        message.metadata = event_receive.Metadata
        message.body = event_receive.Body
        message.sequence = event_receive.Sequence
        message.tags = tags
        return message

    def __repr__(self):
        return f"EventStoreMessageReceived: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, from_client_id={self.from_client_id}, timestamp={self.timestamp}, sequence={self.sequence}, tags={self.tags}"



import logging
import time
from kubemq.cq import *

def main():

    # try:
        client = Client(address="localhost:50000")
        def on_receive_query(request: QueryMessageReceived):
            try:
                print(f"Id:{request.id}, Body:{request.body.decode('utf-8')}")
                response = QueryResponseMessage(
                    query_received=request,
                    is_executed=True,
                    body=b"hello kubemq, I'm replying to you!",
                )
                client.send_response_message(response)
            except Exception as e:
                print(e)


        def on_error_handler(err: str):
            print(f"Error: {err}")

        client.subscribe_to_queries(
            subscription=QueriesSubscription(
                channel="q1",
                group="",
                on_receive_query_callback=on_receive_query,
                on_error_callback=on_error_handler,
            )
            , cancel=CancellationToken())
        time.sleep(1)
        response: QueryResponseMessage = client.send_query_request(QueryMessage(
            channel="q1",
            body=b"hello kubemq, please reply to me!",
            timeout_in_seconds=10,
        ))
        print(
            f"Request Execution: {response.is_executed}, Body: {response.body} Executed at: {response.timestamp}, Error: {response.error}")


    # except Exception as e:
    #         print(e)
    #         return


if __name__ == "__main__":
    main()



import uuid
from typing import Dict
from kubemq.grpc import QueueMessage as pbQueueMessage
from kubemq.grpc import QueuesUpstreamRequest as pbQueuesUpstreamRequest
from kubemq.grpc import QueueMessagePolicy as pbQueueMessagePolicy


class QueueMessage:
    """A class representing a message in a queue.

    Attributes:
        id (str): The unique identifier for the message.
        channel (str): The channel of the message.
        metadata (str): The metadata associated with the message.
        body (bytes): The body of the message.
        tags (Dict[str, str]): The tags associated with the message.
        delay_in_seconds (int): The delay in seconds before the message becomes available in the queue.
        expiration_in_seconds (int): The expiration time in seconds for the message.
        attempts_before_dead_letter_queue (int): The number of receive attempts allowed for the message before it is moved to the dead letter queue.
        dead_letter_queue (str): The dead letter queue where the message will be moved after reaching the maximum receive attempts.

    Methods:
        validate(): Validates the message attributes and ensures that the required attributes are set.
        encode(client_id: str) -> pbQueuesUpstreamRequest: Encodes the message into a protocol buffer format for sending to the queue server.
        encode_message(client_id: str) -> pbQueueMessage: Encodes the message into a protocol buffer format.
    """
    def __init__(self, id: str = None,
                 channel: str = None,
                 metadata: str = None,
                 body: bytes = b"",
                 tags: Dict[str, str] = None,
                 delay_in_seconds: int = 0,
                 expiration_in_seconds: int = 0,
                 attempts_before_dead_letter_queue: int = 0,
                 dead_letter_queue: str = "",
                 ):
        self.id: str = id
        self.channel: str = channel
        self.metadata: str = metadata
        self.body: bytes = body
        self.tags: Dict[str, str] = tags if tags else {}
        self.delay_in_seconds: int = delay_in_seconds
        self.expiration_in_seconds: int = expiration_in_seconds
        self.attempts_before_dead_letter_queue: int = attempts_before_dead_letter_queue
        self.dead_letter_queue: str = dead_letter_queue

    def validate(self) -> 'QueueMessage':
        if not self.channel:
            raise ValueError("Queue message must have a channel.")

        if not self.metadata and not self.body and not self.tags:
            raise ValueError("Queue message must have at least one of the following: metadata, body, or tags.")

        if self.attempts_before_dead_letter_queue < 0:
            raise ValueError("Queue message attempts_before_dead_letter_queue must be a positive number.")

        if self.delay_in_seconds < 0:
            raise ValueError("Queue message delay_in_seconds must be a positive number.")

        if self.expiration_in_seconds < 0:
            raise ValueError("Queue message expiration_in_seconds must be a positive number.")

        return self

    def encode(self, client_id: str) -> pbQueuesUpstreamRequest:
        pb_queue_stream = pbQueuesUpstreamRequest()
        pb_queue_stream.RequestID = str(uuid.uuid4())
        pb_message = self.encode_message(client_id)
        pb_queue_stream.Messages.append(pb_message)
        return pb_queue_stream

    def encode_message(self, client_id: str) -> pbQueueMessage:
        pb_queue = pbQueueMessage()  # Assuming pb is an imported module
        pb_queue.MessageID = self.id or str(uuid.uuid4())
        pb_queue.ClientID = client_id
        pb_queue.Channel = self.channel
        pb_queue.Metadata = self.metadata or ""
        pb_queue.Body = self.body
        pb_queue.Tags.update(self.tags)
        pb_queue.Policy.DelaySeconds = self.delay_in_seconds or 0
        pb_queue.Policy.ExpirationSeconds = self.expiration_in_seconds or 0
        pb_queue.Policy.MaxReceiveCount = self.attempts_before_dead_letter_queue or 0
        pb_queue.Policy.MaxReceiveQueue = self.dead_letter_queue or ""
        return pb_queue

    def __repr__(self):
        return f"QueueMessage: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, tags={self.tags}, delay_in_seconds={self.delay_in_seconds}, expiration_in_seconds={self.expiration_in_seconds}, attempts_before_dead_letter_queue={self.attempts_before_dead_letter_queue}, dead_letter_queue={self.dead_letter_queue}"


import uuid
from typing import Dict, Callable
from datetime import datetime
from kubemq.grpc import QueueMessage as pbQueueMessage
from kubemq.grpc import QueuesDownstreamRequest, QueuesDownstreamRequestType, QueuesDownstreamResponse


class QueueMessageReceived:
    """
    Class representing a received queue message.

    Attributes:
        id (str): Unique identifier of the message.
        channel (str): Name of the channel the message belongs to.
        metadata (str): Additional metadata associated with the message.
        body (bytes): The content of the message in bytes.
        from_client_id (str): Identifier of the client who sent the message.
        tags (Dict[str, str]): Dictionary of key-value pairs representing tags associated with the message.
        timestamp (datetime): Timestamp indicating when the message was received.
        sequence (int): Sequential number indicating the order in which the message was received.
        receive_count (int): Number of times the message has been received.
        is_re_routed (bool): Flag indicating whether the message has been re-routed from another queue.
        re_route_from_queue (str): Name of the queue the message was re-routed from.
        expired_at (datetime): Timestamp indicating when the message will expire.
        delayed_to (datetime): Timestamp indicating when the message will be available for processing.
        transaction_id (str): Identifier of the transaction the message belongs to.
        is_transaction_completed (bool): Flag indicating whether the transaction is completed.
        response_handler (Callable[[QueuesDownstreamRequest], QueuesDownstreamResponse]): Callback function for sending downstream requests.
        receiver_client_id (str): Identifier of the client who received the message.

    Methods:
        ack():
            Sends an acknowledgment for the message.
            Raises a ValueError if the transaction is already completed.

        reject():
            Rejects the message.
            Raises a ValueError if the transaction is already completed or if the rejection fails.

        re_queue(channel: str):
            Re-queues the message to the specified channel.
            Raises a ValueError if the re-queue channel is empty or if the re-queue fails.

        decode(message: pbQueueMessage, transaction_id: str, transaction_is_completed: bool,
               receiver_client_id: str, response_handler: Callable[[QueuesDownstreamRequest], QueuesDownstreamResponse]) -> 'QueueMessageReceived':
            Decodes the given `pbQueueMessage` protobuf message and initialize the instance attributes.
            Returns the updated `QueueMessageReceived` object.

        __repr__() -> str:
            Returns a string representation of the `QueueMessageReceived` object.
    """
    def __init__(self):
        self.id: str = ""
        self.channel: str = ""
        self.metadata: str = ""
        self.body: bytes = b""
        self.from_client_id = ""
        self.tags: Dict[str, str] = {}
        self.timestamp: datetime = datetime.fromtimestamp(0)
        self.sequence: int = 0
        self.receive_count: int = 0
        self.is_re_routed: bool = False
        self.re_route_from_queue: str = ""
        self.expired_at: datetime = datetime.fromtimestamp(0)
        self.delayed_to: datetime = datetime.fromtimestamp(0)
        self.transaction_id: str = ""
        self.is_transaction_completed: bool = False
        self.response_handler: Callable[[QueuesDownstreamRequest], QueuesDownstreamResponse]
        self.receiver_client_id: str = ""

    def ack(self):
        if self.is_transaction_completed:
            raise ValueError("transaction is already completed")
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.receiver_client_id
        request.Channel = self.channel
        request.RequestTypeData = QueuesDownstreamRequestType.AckRange
        request.RefTransactionId = self.transaction_id
        request.SequenceRange.append(self.sequence)
        self.response_handler(request)

    def reject(self):
        if self.is_transaction_completed:
            raise ValueError("transaction is already completed")
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.receiver_client_id
        request.Channel = self.channel
        request.RequestTypeData = QueuesDownstreamRequestType.NAckRange
        request.RefTransactionId = self.transaction_id
        request.SequenceRange.append(self.sequence)
        result = self.response_handler(request)
        if result.IsError:
            raise ValueError(result.Error)

    def re_queue(self, channel: str):
        if not channel:
            raise ValueError("re-queue channel cannot be empty")
        request = QueuesDownstreamRequest()
        request.RequestID = str(uuid.uuid4())
        request.ClientID = self.receiver_client_id
        request.Channel = self.channel
        request.RequestTypeData = QueuesDownstreamRequestType.ReQueueRange
        request.RefTransactionId = self.transaction_id
        request.SequenceRange.append(self.sequence)
        request.ReQueueChannel = channel
        result = self.response_handler(request)
        if result.IsError:
            raise ValueError(result.Error)

    def decode(self, message: pbQueueMessage,
               transaction_id: str,
               transaction_is_completed: bool,
               receiver_client_id: str,
               response_handler: Callable[
                   [QueuesDownstreamRequest], QueuesDownstreamResponse]) -> 'QueueMessageReceived':
        self.id = message.MessageID
        self.channel = message.Channel
        self.metadata = message.Metadata
        self.body = message.Body
        self.from_client_id = message.ClientID
        for tag in message.Tags:
            self.tags[tag] = message.Tags[tag]
        if message.Attributes:
            self.timestamp = datetime.fromtimestamp(message.Attributes.Timestamp / 1e9)
            self.sequence = message.Attributes.Sequence
            self.receive_count = message.Attributes.ReceiveCount
            self.is_re_routed = message.Attributes.ReRouted
            self.re_route_from_queue = message.Attributes.ReRoutedFromQueue
            self.expired_at = datetime.fromtimestamp(message.Attributes.ExpirationAt / 1e6)
            self.delayed_to = datetime.fromtimestamp(message.Attributes.DelayedTo / 1e6)
        self.transaction_id = transaction_id
        self.is_transaction_completed = transaction_is_completed
        self.receiver_client_id = receiver_client_id
        self.response_handler = response_handler
        return self

    def __repr__(self):
        return f"QueueMessageReceived: id={self.id}, channel={self.channel}, metadata={self.metadata}, body={self.body}, from_client_id={self.from_client_id}, timestamp={self.timestamp}, sequence={self.sequence}, receive_count={self.receive_count}, is_re_routed={self.is_re_routed}, re_route_from_queue={self.re_route_from_queue}, expired_at={self.expired_at}, delayed_to={self.delayed_to}, transaction_id={self.transaction_id}, is_transaction_completed={self.is_transaction_completed}, tags={self.tags}"
